@@ -11,7 +11,7 @@ logger = logging.getLogger(__name__)
 
 class fast_pgn_metadata():
     def __init__(self) -> None:
-        self.frames = {}
+        self.frames: dict[int, bytes] = {}
         self.payload_length = 0
         self.bytes_stored = 0
         self.sequence_counter = -1
@@ -33,7 +33,7 @@ class NMEA2000Decoder():
                  dump_pgns:list[int | str]=[],
                  build_network_map: bool = False,
                  ) -> None:
-        self.data = {}
+        self.data: dict[str, fast_pgn_metadata] = {}
         self.dump_TextIOWrapper = None
         self.build_network_map = build_network_map
         self.started_at = datetime.now()
@@ -90,7 +90,7 @@ class NMEA2000Decoder():
                 raise ValueError(f"Invalid PGN type: {type(pgn)}. Must be int or str.")
         return int_list, str_list
 
-    def _decode_fast_message(self, pgn, priority, src, dest, timestamp, can_data, source_iso_name: IsoName | None) -> NMEA2000Message | None:
+    def _decode_fast_message(self, pgn, priority, src, dest, timestamp, can_data: bytes, source_iso_name: IsoName | None, raw_can_data: bytes | str) -> NMEA2000Message | None:
         """Parse a fast packet message and store the data until all frames are received."""
         fast_packet_key = f"{pgn}_{src}_{dest}"
         
@@ -167,7 +167,7 @@ class NMEA2000Decoder():
             nmea = None
             if combined_payload is not None:
                 logger.debug(f"Combined Payload (hex): {combined_payload})")
-                nmea = self._call_decode_function(pgn, priority, src, dest, timestamp, combined_payload, source_iso_name)
+                nmea = self._call_decode_function(pgn, priority, src, dest, timestamp, combined_payload, source_iso_name, raw_can_data)
 
             # Reset the structure for this PGN
             del self.data[fast_packet_key]
@@ -211,7 +211,7 @@ class NMEA2000Decoder():
         # Log the extracted information
         logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {src}, PGN: {pgn}, CAN Data: {reversed_bytes}")
         
-        return self._decode(pgn, priority, src, dest, timestamp, bytes(reversed_bytes), True)
+        return self._decode(pgn, priority, src, dest, timestamp, bytes(reversed_bytes), bytes_data, True)
 
     def decode_yacht_devices_string(self, yd_string: str) -> NMEA2000Message | None:
         """Process an Yacht Devices string and extract the PGN, source ID, and CAN data. Based on: https://www.yachtd.com/downloads/ydwg02.pdf page 62-63"""
@@ -238,7 +238,7 @@ class NMEA2000Decoder():
         # Log the extracted information
         logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {source_id}, PGN: {pgn_id}, CAN Data: {can_data_bytes}")
         
-        return self._decode(pgn_id, priority, source_id, dest, timestamp, bytes(can_data_bytes))
+        return self._decode(pgn_id, priority, source_id, dest, timestamp, bytes(can_data_bytes), yd_string)
 
     def decode_basic_string(self, basic_string: str, already_combined: bool = False) -> NMEA2000Message | None:
         """Process an Actisense packet string and extract the PGN, source ID, and CAN data."""
@@ -265,7 +265,7 @@ class NMEA2000Decoder():
         logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {src}, PGN: {pgn_id}, CAN Data: {can_data_bytes}")
         
         # not calling _decode as in this format the fast frames are already combined
-        return self._decode(pgn_id, priority, src, dest, timestamp, bytes(can_data_bytes), already_combined)
+        return self._decode(pgn_id, priority, src, dest, timestamp, bytes(can_data_bytes), basic_string,already_combined)
 
     @staticmethod
     def _extract_header(frame_id_int: int) -> Tuple[int, int, int, int]:
@@ -317,7 +317,7 @@ class NMEA2000Decoder():
             can_data,
             source_id)
         
-        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data))
+        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data), packet)
 
     def decode_usb(self, packet: bytes) -> NMEA2000Message | None:
         """Tested with Waveshare-usb-a device. Process a single packet and extract the PGN, source ID, and CAN data."""
@@ -349,7 +349,7 @@ class NMEA2000Decoder():
             can_data,
             source_id)
         
-        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data))
+        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data), packet)
 
     @staticmethod
     def _isFastPGN(pgn_id: int) -> bool | None:
@@ -369,7 +369,7 @@ class NMEA2000Decoder():
             logger.warning("Not supporrted PGN: %d", pgn_id)
             self.logged_unsupported_pgns.add(pgn_id)
 
-    def _decode(self, pgn: int, priority: int, source_id: int, destination_id: int, timestamp: datetime, can_data: bytes, already_combined: bool = False) -> NMEA2000Message | None:
+    def _decode(self, pgn: int, priority: int, source_id: int, destination_id: int, timestamp: datetime, can_data: bytes, raw_can_data: bytes | str, already_combined: bool = False) -> NMEA2000Message | None:
         """Decode a single PGN message."""
 
         source_iso_name = None
@@ -407,11 +407,11 @@ class NMEA2000Decoder():
             return None
         
         if is_fast:
-            return self._decode_fast_message(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name)
+            return self._decode_fast_message(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name, raw_can_data)
         else:
-            return self._call_decode_function(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name)
+            return self._call_decode_function(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name, raw_can_data)
 
-    def _call_decode_function(self, pgn:int, priority: int, src: int, dest: int, timestamp: datetime, data:bytes, source_iso_name: IsoName | None) -> NMEA2000Message | None:
+    def _call_decode_function(self, pgn:int, priority: int, src: int, dest: int, timestamp: datetime, data:bytes, source_iso_name: IsoName | None, raw_can_data: bytes | str) -> NMEA2000Message | None:
         decode_func_name = f'decode_pgn_{pgn}'
         decode_func = globals().get(decode_func_name)
 
@@ -449,14 +449,14 @@ class NMEA2000Decoder():
             logger.debug("Excluding (by include) PGN %d by id: %s", pgn, nmea2000Message.id)
             return None
         
+        nmea2000Message.add_data(src, dest, priority, timestamp, source_iso_name, self.build_network_map, raw_can_data)
+        nmea2000Message.apply_preferred_units(self.preferred_units)
+
         # Handle dump to file
         if (self.dump_TextIOWrapper is not None) and (len(self.dump_include_pgns)+len(self.dump_include_pgns_ids) == 0 or nmea2000Message.PGN in self.dump_include_pgns or nmea2000Message.id in self.dump_include_pgns_ids):
             str = nmea2000Message.to_json() + "\n"
             self.dump_TextIOWrapper.write(str)
                     
-        nmea2000Message.add_data(src, dest, priority, timestamp, source_iso_name, self.build_network_map)
-        nmea2000Message.apply_preferred_units(self.preferred_units)
-
         return nmea2000Message
 
     def close(self):
